% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adasyn.R
\name{step_adasyn}
\alias{step_adasyn}
\alias{tidy.step_adasyn}
\title{Adaptive Synthetic Sampling Approach}
\usage{
step_adasyn(
  recipe,
  ...,
  role = NA,
  trained = FALSE,
  column = NULL,
  over_ratio = 1,
  neighbors = 5,
  skip = TRUE,
  seed = sample.int(10^5, 1),
  id = rand_id("adasyn")
)

\method{tidy}{step_adasyn}(x, ...)
}
\arguments{
\item{recipe}{A recipe object. The step will be added to the
sequence of operations for this recipe.}

\item{...}{One or more selector functions to choose which
variable is used to sample the data. See \code{\link[=selections]{selections()}}
for more details. The selection should result in \emph{single
factor variable}. For the \code{tidy} method, these are not
currently used.}

\item{role}{Not used by this step since no new variables are
created.}

\item{trained}{A logical to indicate if the quantities for
preprocessing have been estimated.}

\item{column}{A character string of the variable name that will
be populated (eventually) by the \code{...} selectors.}

\item{over_ratio}{A numeric value for the ratio of the
majority-to-minority frequencies. The default value (1) means
that all other levels are sampled up to have the same
frequency as the most occurring level. A value of 0.5 would mean
that the minority levels will have (at most) (approximately)
half as many rows than the majority level.}

\item{neighbors}{An integer. Number of nearest neighbours that are used
to generate the new examples of the minority class.}

\item{skip}{A logical. Should the step be skipped when the
recipe is baked by \code{\link[recipes:bake.recipe]{bake.recipe()}}? While all operations are baked
when \code{\link[recipes:prep.recipe]{prep.recipe()}} is run, some operations may not be able to be
conducted on new data (e.g. processing the outcome variable(s)).
Care should be taken when using \code{skip = TRUE} as it may affect
the computations for subsequent operations}

\item{seed}{An integer that will be used as the seed when
applied.}

\item{id}{A character string that is unique to this step to identify it.}

\item{x}{A \code{step_adasyn} object.}
}
\value{
An updated version of \code{recipe} with the new step
added to the sequence of existing steps (if any). For the
\code{tidy} method, a tibble with columns \code{terms} which is
the variable used to sample.
}
\description{
\code{step_adasyn} creates a \emph{specification} of a recipe
step that generates synthetic positive instances using ADASYN algorithm.
}
\details{
All columns in the data are sampled and returned by \code{\link[=juice]{juice()}}
and \code{\link[=bake]{bake()}}.

All columns used in this step must be numeric with no missing data.

When used in modeling, users should strongly consider using the
option \code{skip = TRUE} so that the extra sampling is \emph{not}
conducted outside of the training set.
}
\examples{
library(recipes)
library(modeldata)
data(okc)

sort(table(okc$Class, useNA = "always"))

ds_rec <- recipe(Class ~ age + height, data = okc) \%>\%
  step_meanimpute(all_predictors()) \%>\%
  step_adasyn(Class) \%>\%
  prep()

sort(table(juice(ds_rec)$Class, useNA = "always"))

# since `skip` defaults to TRUE, baking the step has no effect
baked_okc <- bake(ds_rec, new_data = okc)
table(baked_okc$Class, useNA = "always")

library(ggplot2)

ggplot(circle_example, aes(x, y, color = class)) +
  geom_point() +
  labs(title = "Without ADASYN")

recipe(class ~ ., data = circle_example) \%>\%
  step_adasyn(class) \%>\%
  prep() \%>\%
  juice() \%>\%
  ggplot(aes(x, y, color = class)) +
  geom_point() +
  labs(title = "With ADASYN")

}
\references{
He, H., Bai, Y., Garcia, E. and Li, S. 2008. ADASYN: Adaptive
synthetic sampling approach for imbalanced learning. Proceedings of
IJCNN 2008. (IEEE World Congress on Computational Intelligence). IEEE
International Joint Conference. pp.1322-1328.
}
\concept{preprocessing}
\concept{subsampling}
\keyword{datagen}
